////////////////////////////////////////////////////////////////////////
// Class:       BlipReco3D
// Plugin Type: producer (art v2_11_03)
// File:        BlipReco3D_module.cc
//
// W. Foreman
// May 2022
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/View.h"
#include "canvas/Persistency/Common/FindMany.h"
#include "canvas/Persistency/Common/FindManyP.h"
#include "canvas/Persistency/Common/FindOneP.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// LArSoft includes
#include "larcore/Geometry/Geometry.h"
#include "lardata/Utilities/AssociationUtil.h"
#include "lardata/Utilities/GeometryUtilities.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardata/DetectorInfoServices/LArPropertiesService.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "lardataobj/Simulation/SimEnergyDeposit.h"
#include "lardataobj/Simulation/AuxDetSimChannel.h"
#include "lardataobj/AnalysisBase/Calorimetry.h"
#include "lardataobj/AnalysisBase/ParticleID.h"
#include "lardataobj/AnalysisBase/CosmicTag.h"
#include "lardataobj/AnalysisBase/BackTrackerMatchingData.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Track.h"
#include "lardataobj/RecoBase/Shower.h"
#include "lardataobj/RecoBase/Cluster.h"
#include "lardataobj/RecoBase/SpacePoint.h"
#include "lardataobj/RecoBase/Hit.h"
#include "lardataobj/RecoBase/EndPoint2D.h"
#include "lardataobj/RecoBase/Vertex.h"
#include "larcoreobj/SimpleTypesAndConstants/geo_types.h"
#include "larreco/Calorimetry/CalorimetryAlg.h"
#include "cetlib/search_path.h"

// MicroBooNE-specific includes
#include "ubevt/Database/UbooneElectronLifetimeProvider.h"
#include "ubevt/Database/UbooneElectronLifetimeService.h"
#include "ubreco/BlipReco/Alg/BlipRecoAlg.h"

class BlipReco3D;

class BlipReco3D : public art::EDProducer 
{
  public:
  explicit BlipReco3D(fhicl::ParameterSet const & p);
  BlipReco3D(BlipReco3D const &) = delete;
  BlipReco3D(BlipReco3D &&) = delete;
  BlipReco3D & operator = (BlipReco3D const &) = delete;
  BlipReco3D & operator = (BlipReco3D &&) = delete;

  // Required functions.
  void produce(art::Event & e) override;

  private:
  blip::BlipRecoAlg       fBlipAlg;

};



//###################################################
//  BlipReco3D constructor and destructor
//###################################################
BlipReco3D::BlipReco3D(fhicl::ParameterSet const & pset) : art::EDProducer(pset)
  , fBlipAlg ( pset.get<fhicl::ParameterSet>("BlipAlg") )
{
  // Read in fcl parameters for blip reco alg
  fhicl::ParameterSet pset_blipalg = pset.get<fhicl::ParameterSet>("BlipAlg");
 
  // produce spacepoints and 'hit <--> spacepoint' associations
  produces< std::vector<  recob::SpacePoint > >();
  produces< art::Assns <  recob::Hit, recob::SpacePoint> >();
  
  // produce blips and 'hit <--> blip' associations
  produces< std::vector<  blipobj::Blip > >();
  produces< art::Assns <  recob::Hit, blipobj::Blip> >();
  
}



//###################################################
//  Main produce function
//###################################################
void BlipReco3D::produce(art::Event & evt)
{
  
  std::cout<<"\n"
  <<"------- BlipRecoProducer --------------\n"
  <<"Event "<<evt.id().event()<<" / run "<<evt.id().run()<<"\n";

  int blipCount = 0;

  //============================================
  // Make unique pointers to the vectors of objects 
  // and associations we will create
  //============================================
  std::unique_ptr< std::vector< recob::SpacePoint> > SpacePoint_v(new std::vector<recob::SpacePoint>);
  std::unique_ptr< art::Assns <recob::Hit, recob::SpacePoint> >  assn_hit_sps_v(new art::Assns<recob::Hit,recob::SpacePoint> );

  std::unique_ptr< std::vector< blipobj::Blip > > Blip_v(new std::vector<blipobj::Blip>);
  std::unique_ptr< art::Assns <recob::Hit, blipobj::Blip> >  assn_hit_blip_v(new art::Assns<recob::Hit,blipobj::Blip> );
 
  
  //============================================
  // Run blip reconstruction: 
  //============================================
  fBlipAlg.RunBlipReco(evt);

  //============================================
  // Get hits from input module
  //============================================
  auto hitprod = fBlipAlg.fHitProducer;
  std::cout<<"Input hits: "<<hitprod.label()<<"::"<<hitprod.process()<<"\n";
  art::Handle< std::vector<recob::Hit> > hitHandle;
  std::vector<art::Ptr<recob::Hit> > hitlist;
  if (evt.getByLabel(hitprod,hitHandle))
    art::fill_ptr_vector(hitlist, hitHandle);
  std::cout<<"Retrieved "<<hitlist.size()<<" hits from event\n"; 

  //===========================================
  // Make recob::SpacePoints and objects
  //===========================================
  for(size_t i=0; i<fBlipAlg.blips.size(); i++){
    auto& b = fBlipAlg.blips[i];
    
    if( !b.isValid ) continue;
    
    Blip_v->emplace_back(b);
    blipCount++;

    Double32_t xyz[3];
    Double32_t xyz_err[6];
    Double32_t chiSquare = 0;
    Double32_t err = 0.; //b.MaxIntersectDiff?
    xyz[0]      = b.Position.X();
    xyz[1]      = b.Position.Y();
    xyz[2]      = b.Position.Z();
    xyz_err[0]  = err;
    xyz_err[1]  = err;
    xyz_err[2]  = err;
    xyz_err[3]  = err;
    xyz_err[4]  = err;
    xyz_err[5]  = err;
    
    recob::SpacePoint newpt(xyz,xyz_err,chiSquare);
    SpacePoint_v->emplace_back(newpt);
    
    // Hit associations 
    for(auto& hc : b.clusters ) {
      for(auto& ihit : hc.HitIDs ) {
        auto& hitptr = hitlist[ihit];
        util::CreateAssn(*this, evt, *SpacePoint_v, hitptr, *assn_hit_sps_v);
        util::CreateAssn(*this, evt, *Blip_v,       hitptr, *assn_hit_blip_v); 
      }
    }
  }
    
  //===========================================
  // Put them on the event
  //===========================================
  evt.put(std::move(SpacePoint_v));
  evt.put(std::move(assn_hit_sps_v));
  
  evt.put(std::move(Blip_v));
  evt.put(std::move(assn_hit_blip_v));
  
  std::cout
  <<"Added "<<blipCount<<" 3D blips to the event.\n"
  <<"---------------------------------------\n";

}//END EVENT LOOP

DEFINE_ART_MODULE(BlipReco3D)
